/*
 *	Serial Line interface
 *
 *	Rick Adams
 *	Center for Seismic Studies
 *	1300 N 17th Street, Suite 1450
 *	Arlington, Virginia 22209
 *	(703)276-7900
 *	rick@seismo.ARPA
 *	seismo!rick
 *
 *	Some things done here could obviously be done in a better way,
 *	but they were done this way to minimize the number of files
 *	that had to be changed to accomodate this device.
 *	A lot of this code belongs in the tty driver.
 */

static char rcsid[] = "$Header: if_sl.c,v 1.11 84/10/04 12:54:47 rick Exp $";

#include "sl.h"
#if NSL > 0

#include "../h/param.h"
#include "../h/mbuf.h"
#include "../h/buf.h"
#include "../h/socket.h"
#include "../h/ioctl.h"
#include "../h/tty.h"
#include "../h/errno.h"

#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"
#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"

#ifdef vax
#include "../vax/mtpr.h"
#endif vax

struct sl_softc {
	struct	ifnet sc_if;		/* network-visible interface */
	short	sc_escaped;		/* flags */
	short	sc_oactive;		/* active or not */
	struct	tty *sc_ttyp;		/* pointer to tty structure */
	int	sc_ttyspeed;		/* baud rate of line */
	int	sc_len;			/* length of buffer remaining */
	struct mbuf *sc_m;		/* head of mbuf chain */
	struct mbuf *sc_mt;		/* tail of mbuf chain */
	u_char *sc_mp;			/* pointer to next available buf char */
}sl_softc[NSL];

#define FRAME_END	 	0300		/* Frame End */
#define FRAME_ESCAPE		0333		/* Frame Esc */
#define TRANS_FRAME_END	 	0334		/* transposed frame end */
#define TRANS_FRAME_ESCAPE 	0335		/* transposed frame esc */

#define SLTU	1006

#ifdef sun
#define T_LINEP	t_nlines
#endif

int sloutput(), slioctl();

slopen(dev,tp)
dev_t dev;
register struct tty *tp;
{
	register struct sl_softc *sc = sl_softc;
	register struct sockaddr_in *sin;
	register int nsl;
	extern int ifqmaxlen;
	int addloop = 0;

	if (tp->T_LINEP)
		return EBUSY;

	for(nsl = 0; sc->sc_ttyp; sc++)
		if(nsl++ >= NSL)
			return ENOSPC;

	/* This is a real kludge. Since we do the attach in the rc file,
	 * the loopback driver gets configured first. Now Berkeley, in its
	 * infinite wisdom, when connecting to a machine through a gateway,
	 * chooses the FIRST interface on its list. Guess what that is when you
	 * don't have an ethernet? That's right! It's the loopback driver!
	 * The following crock is an attempt to get around that.
	 */

	if (ifnet && ifnet->if_name[0] == 'l' && ifnet->if_name[1] == 'o'
		&& ifnet->if_name[2] == '\0') {
		/* the loopback device is configured in first */
		ifnet = ifnet->if_next;
		addloop++;
	}

	bzero((caddr_t)sc, sizeof(struct sl_softc));
#ifdef sun
	tp->T_LINEP = (int)sc;
#else vax
	tp->T_LINEP = (caddr_t)sc;
#endif vax
	sc->sc_ttyp = tp;
	sc->sc_if.if_unit = nsl;
	sc->sc_if.if_name = "sl";
	sc->sc_if.if_mtu = SLTU;
	sc->sc_if.if_output = sloutput;
	sc->sc_if.if_ioctl = slioctl;
	sc->sc_if.if_flags = IFF_POINTOPOINT;
	sc->sc_if.if_snd.ifq_maxlen = ifqmaxlen;
	if_attach(&sc->sc_if);

	if (addloop)
		loattach();

	return 0;
}

/*ARGSUSED*/
sltioctl(tp, cmd, data, flag)
struct tty *tp;
caddr_t data;
{
	register struct sl_softc *sc = (struct sl_softc *)tp->T_LINEP;

	/*
	 * This is done this way, because we want these atributes to
	 * be present AFTER the user program closes the line.
	 */

	if (cmd == TIOCGETD){
		*(int *)data = (int)sc->sc_if.if_unit;
		tp->t_state &= ~TS_HUPCLS;
		tp->t_flags |= NOHANG;
		sc->sc_ttyspeed = tp->t_ispeed;
		return 0;
	}
	return EINVAL;
}
sloutput(ifp, m, dst)
register struct ifnet *ifp;
register struct mbuf *m;
struct sockaddr *dst;
{
	int s, looutput();
	extern struct ifnet loif;

	if (dst->sa_family != AF_INET) {
		printf("sl%d: af%d not supported\n", ifp->if_unit,
			dst->sa_family);
		m_freem(m);
		return EAFNOSUPPORT;
	}
#ifdef doesnt_work
	/* if it's for me, use the loopback driver */
	if (bcmp(dst->sa_data, ifp->if_addr.sa_data, 14) == 0)
		return looutput(&loif, m, dst);
#endif

	s = splimp();
	if (IF_QFULL(&ifp->if_snd)) {
		IF_DROP(&ifp->if_snd);
		m_freem(m);
		splx(s);
		sl_softc[ifp->if_unit].sc_if.if_collisions++;
		return ENOBUFS;
	}
	IF_ENQUEUE(&ifp->if_snd, m);
	if (sl_softc[ifp->if_unit].sc_oactive == 0)
		return slstart(ifp->if_unit, s); /* slstart does an splx(s) */

	splx(s);
	return 0;
}

/*
 * Start output on interface.  Get another datagram
 * to send from the interface queue and map it to
 * the interface before starting output.
 */
slstart(n, s)
int n, s;
{
	register struct sl_softc *sc = &sl_softc[n];
	register struct tty *tp;
	struct mbuf *m, *mp;

	IF_DEQUEUE(&sc->sc_if.if_snd, m);
	splx(s);
	if (m == 0)
		return 0;
	tp = sc->sc_ttyp;
	if (tp == NULL){
		printf("sl: sc_ttyp is 0\n");
		m_freem(m);
		return EHOSTUNREACH;
	}
	sc->sc_oactive = 1;

	for (mp = m; mp; mp = mp->m_next) {
		register unsigned len = mp->m_len;
		register u_char *mcp;
		u_char c;

		mcp = mtod(mp, u_char *);
		while (len--){
			c = *mcp++;
			if( c == FRAME_ESCAPE || c == FRAME_END) {
				if (putc(FRAME_ESCAPE, &tp->t_outq))
					goto full;
				c = c == FRAME_ESCAPE ?
					TRANS_FRAME_ESCAPE : TRANS_FRAME_END;
			}
			if (putc(c, &tp->t_outq))
				goto full;
		}
	}

	if (putc(FRAME_END, &tp->t_outq)){
		/*
		 * if you get many oerrors (more than 1 or two a day)
		 * you probably do not have enough clists and you should 
		 * increase "nclist" in param.c
		 */
full:
		m_freem(m);
		(void) unputc(&tp->t_outq);	/* make some space */
		putc(FRAME_END, &tp->t_outq);	/* end the packet */
		sc->sc_if.if_oerrors++;
		sc->sc_oactive = 0;
		ttstart(tp);
		return ENOBUFS;
	}
	m_freem(m);
	ttstart(tp);
	sc->sc_oactive = 0;
	sc->sc_if.if_opackets++;
	return 0;
}

/*
 * tty interface receiver interrupt.
 */
slinput(c, tp)
register c;
register struct tty *tp;
{
	register struct sl_softc *sc;
	struct mbuf *m;
	register struct ifqueue *inq;
	int s;

	sc = (struct sl_softc *)tp->T_LINEP;
	if (sc == NULL){
		printf("T_LINEP NULL\n");
		return;
	}
	c &= 0xff;
	if (sc->sc_escaped) {
		sc->sc_escaped = 0;
		switch(c) {
		case TRANS_FRAME_ESCAPE:
			c = FRAME_ESCAPE;
			break;
		case TRANS_FRAME_END:
			c = FRAME_END;
			break;
		default:
			sc->sc_if.if_ierrors++;
			return;
		}
	} else {
		switch(c) {
		case FRAME_END:
			if (sc->sc_mt == 0) {
				sc->sc_if.if_ierrors++;
				return;
			}
			sc->sc_if.if_ipackets++;
			sc->sc_mt->m_len -= sc->sc_len;
			schednetisr(NETISR_IP);
			inq = &ipintrq;

			s = splimp();
			if (IF_QFULL(inq)) {
				IF_DROP(inq);
				sc->sc_if.if_ierrors++;
				m_freem(sc->sc_m);
			} else
				IF_ENQUEUE(inq, sc->sc_m);
			splx(s);
			sc->sc_mt = 0;
			sc->sc_len = 0;
			return;
		case FRAME_ESCAPE:
			sc->sc_escaped = 1;
			return;
		}
	}
	if (sc->sc_len <= 0) {	/* have to get more buffer space */
		register struct mbuf *mm;

		MGET(mm, M_DONTWAIT, MT_DATA);
		if (mm == 0) {
			m_freem(sc->sc_m);
			sc->sc_mt = 0;
			sc->sc_len = 0;
			sc->sc_if.if_collisions++;
			return;
		}
		if (sc->sc_mt == 0){
			sc->sc_mt = sc->sc_m = mm;
		} else {
			sc->sc_mt->m_next = mm;
			sc->sc_mt = mm;
		}
		sc->sc_len = mm->m_len = MLEN;
		sc->sc_mp = mtod(mm, u_char *);
	}

	*sc->sc_mp++ = c;
	sc->sc_len--;
}

/*
 * Process an ioctl request.
 */
slioctl(ifp, cmd, data)
register struct ifnet *ifp;
int cmd;
caddr_t data;
{
	struct ifreq *ifr = (struct ifreq *)data;
	struct tty *tp;
	struct sockaddr_in *sin;
	struct sl_softc *sc;
	struct rtentry route;
#ifdef sun
	struct sockaddr_in SIN;
#endif sun
	int s = splimp(), error = 0;

#ifdef vax
# include "dh.h"
# if NDH > 0
	int dhstart();
# endif NDH
# include "dz.h"
# if NDZ > 0
	int dzstart();
# endif NDZ
# include "dmf.h"
# if NDMF > 0
	int dmfstart();
# endif NDMF
#endif vax

#ifdef sun
# include "zs.h"
# if NZS > 0
	int zsstart();
# endif  NZS
#endif sun

	switch (cmd) {
	case SIOCSIFADDR:
		if (ifp->if_flags & IFF_RUNNING)
			if_rtinit(ifp, -1);	/* delete previous route */
#ifdef sun
		ifp->if_addr = *(struct sockaddr *)data;
#else vax
		sin = (struct sockaddr_in *)&ifr->ifr_addr;
		ifp->if_addr = *(struct sockaddr *)sin;
#endif vax
		sin = (struct sockaddr_in *)&ifp->if_dstaddr;
		ifp->if_net = in_netof(sin->sin_addr);
		ifp->if_flags |= IFF_UP | IFF_RUNNING;
		/* set up routing table entry */
		if ((ifp->if_flags & IFF_ROUTE) == 0) {
			rtinit(&ifp->if_dstaddr, &ifp->if_addr, RTF_HOST|RTF_UP);
			ifp->if_flags |= IFF_ROUTE;
		}
		/* this really belongs elsewhere, but we don't want it reset
		 * when the file descriptor is closed
		 */
		sc = &sl_softc[ifp->if_unit];
		tp = sc->sc_ttyp;
		tp->t_line = SLIPDISC;
		tp->t_flags = RAW|EVENP|ODDP;
		tp->t_ispeed = tp->t_ospeed = sc->sc_ttyspeed;

		/* Unfortunately, the tty driver does not have any
		 * way of jumping directly to the correct param routine
		 * so, to get the line set to the correct baudrate
		 * we are forced to use the following kludge
		 * (Fortunately, it is only called once per line)
		 */
#ifdef vax
# if NDH > 0
		if (tp->t_oproc == dhstart)
			dhparam(minor(tp->t_dev));
		else
# endif NDH
# if NDZ > 0
		if (tp->t_oproc == dzstart)
			dzparam(minor(tp->t_dev));
		else
# endif NDZ
# if NDMF > 0
		if (tp->t_oproc == dmfstart)
			dmfparam(minor(tp->t_dev));
		else
# endif NDMF
#endif vax

#ifdef sun
# if NZS > 0
		if (tp->t_oproc == zsstart)
			zsparam(minor(tp->t_dev));
		else
# endif NZS
#endif sun
		{
			printf("Serial line attach: Unknown device\n");
			error = ENODEV;
		}
		break;

	case SIOCSIFDSTADDR:
#ifdef sun
		/* This is horrible, but they leave me little choice */
		bzero((caddr_t)&SIN, sizeof(struct sockaddr_in));
		SIN.sin_addr.s_addr = (u_long)data;
		SIN.sin_family = AF_INET; /* so much for other families... */
		sin = &SIN;
		ifp->if_dstaddr = *(struct sockaddr *)sin;
#else vax
		ifp->if_dstaddr = ifr->ifr_dstaddr;
#endif vax
		break;
#ifdef SIOCIFDETACH
	case SIOCIFDETACH:
		bzero((caddr_t)&route, sizeof (route));
		route.rt_dst = ifp->if_dstaddr;
		route.rt_gateway = ifp->if_addr;
		route.rt_flags = RTF_HOST | RTF_UP;
		(void) rtrequest(SIOCDELRT, &route);
		if_detach(ifp);
		sc = &sl_softc[ifp->if_unit];
		tp = sc->sc_ttyp;
		tp->T_LINEP = 0;
		sc->sc_ttyp = 0;
		ttyclose(tp);
		break;

#endif SIOCIFDETACH
	default:
		error = EINVAL;
	}
	splx(s);
	return error;
}
#endif NSL
